@@ -192,22 +192,65 @@ static void print_help(int advanced)
 
 
 
+/*
+ * get_default_project_root() - scan some common directories for the X
+ * project root
+ *
+ * Users of this information should be careful to account for the
+ * modular layout.
+ */
+
+char *get_default_project_root(void)
+{
+    char *paths[] = { "/usr/X11R6", "/usr/X11", NULL };
+    struct stat stat_buf;
+    int i;
+        
+    for (i = 0; paths[i]; i++) {
+        
+        if (stat(paths[i], &stat_buf) == -1) {
+            continue;
+        }
+    
+        if (S_ISDIR(stat_buf.st_mode)) {
+            return paths[i];
+        }
+    }
+    
+    /* default to "/usr/X11R6", I guess */
+
+    return paths[0];
+
+} /* get_default_project_root() */
+
+
+
 /*
  * parse_commandline() - malloc an Options structure, initialize it,
  * and fill in any pertinent data from the commandline arguments
  */
 
-void parse_commandline(Options *op, int argc, char *argv[])
+Options *parse_commandline(int argc, char *argv[])
 {
+    Options *op;
     int c, boolval;
     char *strval;
     int intval, disable;
+    double doubleval;
 
+    op = (Options *) nvalloc(sizeof(Options));
+    
+    op->gop.x_project_root = get_default_project_root();
+    op->nvagp = -1;
+    op->transparent_index = -1;
+    op->stereo = -1;
+    op->cool_bits = -1;
+    op->tv_over_scan = -1.0;
     
     while (1) {
         
         c = nvgetopt(argc, argv, __options, &strval,
-                     &boolval, &intval, &disable);
+                     &boolval, &intval, &doubleval, &disable);
 
         if (c == -1)
             break;
@@ -239,7 +282,8 @@ void parse_commandline(Options *op, int argc, char *argv[])
             if ((op->depth != 8) &&
                 (op->depth != 15) &&
                 (op->depth != 16) &&
-                (op->depth != 24)) {
+                (op->depth != 24) &&
+                (op->depth != 30)) {
                 fprintf(stderr, "\n");
                 fprintf(stderr, "Invalid depth: %d.\n", op->depth);
                 fprintf(stderr, "\n");
@@ -263,6 +307,45 @@ void parse_commandline(Options *op, int argc, char *argv[])
 
         case FORCE_GENERATE_OPTION: op->force_generate = TRUE; break;
 
+        case ACPID_SOCKET_PATH_OPTION: 
+            if (disable) {
+                op->acpid_socket_path = NV_DISABLE_STRING_OPTION;
+            } else {
+                op->acpid_socket_path = strval; 
+            }
+            break;
+
+        case HANDLE_SPECIAL_KEYS_OPTION:
+            {
+                const char *valid_values[] = {
+                    "Always",
+                    "Never",
+                    "WhenNeeded",
+                    NULL,
+                };
+
+                int i;
+
+                if (disable) {
+                    op->handle_special_keys = NV_DISABLE_STRING_OPTION;
+                    break;
+                }
+
+                for (i = 0; valid_values[i]; i++) {
+                    if (!strcasecmp(strval, valid_values[i])) {
+                        break;
+                    }
+                }
+
+                if (valid_values[i]) {
+                    op->handle_special_keys = strval;
+                } else {
+                    fprintf(stderr, "Invalid HandleSpecialKeys option: %s.\n", strval);
+                    goto fail;
+                }
+                break;
+            }
+
         case NVAGP_OPTION:
 
             /* mark as disabled, so we can remove the option later */
@@ -306,6 +389,101 @@ void parse_commandline(Options *op, int argc, char *argv[])
             op->transparent_index = intval;
             break;
 
+        case TV_STANDARD_OPTION:
+            
+            {
+                const char* valid_values[] = {
+                    "PAL-B",
+                    "PAL-D",
+                    "PAL-G",
+                    "PAL-H",
+                    "PAL-I",
+                    "PAL-K1",
+                    "PAL-M",
+                    "PAL-N",
+                    "PAL-NC",
+                    "NTSC-J",
+                    "NTSC-M",
+                    "HD480i",
+                    "HD480p",
+                    "HD720p",
+                    "HD1080i",
+                    "HD1080p",
+                    "HD576i",
+                    "HD576p",
+                    NULL
+                };
+                int i;
+                
+                /* mark as disabled, so we can remove the option later */
+
+                if (disable) {
+                    op->tv_standard = NV_DISABLE_STRING_OPTION;
+                    break;
+                }
+
+                for (i = 0; valid_values[i]; i++) {
+                    if (!strcasecmp(strval, valid_values[i]))
+                        break;
+                }
+
+                if (valid_values[i]) {
+                    op->tv_standard = strval;
+                } else {
+                    fprintf(stderr, "Invalid TVStandard option: %s.\n", strval);
+                    goto fail;
+                }
+            }
+            break;
+
+        case TV_OUT_FORMAT_OPTION:
+            
+            /* mark as disabled, so we can remove the option later */
+
+            if (disable) {
+                op->tv_out_format = NV_DISABLE_STRING_OPTION;
+                break;
+            }
+
+            if (!strcasecmp(strval, "SVIDEO")) {
+                op->tv_out_format = "SVIDEO";
+            } else if (!strcasecmp(strval, "COMPOSITE")) {
+                op->tv_out_format = "COMPOSITE";
+            } else {
+                fprintf(stderr, "Invalid TVOutFormat option: %s.\n", strval);
+                goto fail;
+            }
+            break;
+
+        case TV_OVER_SCAN_OPTION:
+            
+            /* mark as disabled, so we can remove the option later */
+
+            if (disable) {
+                op->tv_over_scan = -2.0;
+                break;
+            }
+
+            if (doubleval >= 0.0 && doubleval <= 1.0) {
+                op->tv_over_scan = doubleval;
+            } else {
+                fprintf(stderr, "Invalid TVOverScan value: %f.\n", doubleval);
+                goto fail;
+            }
+            break;
+
+        case COOL_BITS_OPTION:
+
+            /* mark as disabled, so we can remove the option later */
+
+            if (disable) {
+                op->cool_bits = -2;
+                break;
+            }
+
+            op->cool_bits = intval;
+            break;
+
         case STEREO_OPTION:
 
             /* mark as disabled, so we can remove the option later */
@@ -334,6 +512,24 @@ void parse_commandline(Options *op, int argc, char *argv[])
             }
             break;
 
+        case MODE_LIST_OPTION:
+            {
+                char *token;
+                token = strtok(strval, " ");
+                if (!token) {
+                    fprintf(stderr, "\n");
+                    fprintf(stderr, "Invalid Mode List string: %s.\n", strval);
+                    fprintf(stderr, "\n");
+                    goto fail;
+                }
+                do {
+                    nv_text_rows_append(&op->add_modes_list, token);
+                    token = strtok(NULL, " ");
+                } while (token != NULL);
+
+                break;
+            }
+
         case REMOVE_MODE_OPTION:
             nv_text_rows_append(&op->remove_modes, strval);
             break;
@@ -463,7 +659,7 @@ void parse_commandline(Options *op, int argc, char *argv[])
         case QUERY_GPU_INFO_OPTION: op->query_gpu_info = TRUE; break;
 
         case 'E':
-            op->extract_edids_from_log = strval;
+            op->extract_edids_from_file = strval;
             break;
 
         case EXTRACT_EDIDS_OUTPUT_FILE_OPTION:
@@ -534,6 +730,15 @@ void parse_commandline(Options *op, int argc, char *argv[])
             op->logo_path = disable ? NV_DISABLE_STRING_OPTION : strval;
             break;
   
+        case USE_DISPLAY_DEVICE_OPTION:
+            op->use_display_device = 
+                disable ? NV_DISABLE_STRING_OPTION : strval;
+            break;
+       
+        case CUSTOM_EDID_OPTION:
+            op->custom_edid = disable ? NV_DISABLE_STRING_OPTION : strval;
+            break;
+          
         default:
             goto fail;
         }
@@ -543,40 +748,7 @@ void parse_commandline(Options *op, int argc, char *argv[])
     
     op->xconfig = tilde_expansion(op->xconfig);
     op->output_xconfig = tilde_expansion(op->output_xconfig);
-
-    return;
     
- fail:
-    
-    fprintf(stderr, "\n");
-    fprintf(stderr, "Invalid commandline, please run `%s --help` "
-            "for usage information.\n", argv[0]);
-    fprintf(stderr, "\n");
-    exit(1);
-
-} /* parse_commandline() */
-
-
-
-/*
- * load_default_options - malloc an Options structure
- * and initialize it with default values.
- *
- */
-
-Options *load_default_options(void)
-{
-    Options *op;
-
-    op = (Options *) nvalloc(sizeof(Options));
-    if (!op) return NULL;
-    
-    op->nvagp = -1;
-    op->transparent_index = -1;
-    op->stereo = -1;
-
-    xconfigGenerateLoadDefaultOptions(&op->gop);
-
     /*
      * XXX save the option structure so that printing routines can
      * access it (and so that we don't have to carry the op everywhere
@@ -588,9 +760,20 @@ Options *load_default_options(void)
         __op = op;
     }
     
-    return op;
+    return (op);
+    
+ fail:
+    
+    fprintf(stderr, "\n");
+    fprintf(stderr, "Invalid commandline, please run `%s --help` "
+            "for usage information.\n", argv[0]);
+    fprintf(stderr, "\n");
+    exit(1);
+    
+    return NULL;
+
+} /* parse_commandline() */
 
-} /* load_default_options() */
 
 
 
@@ -902,6 +1085,8 @@ int update_xconfig(Options *op, XConfigPtr config)
 
     update_modules(config);
 
+    update_server_flags(op, config);
+
     update_banner(config);
     
     return TRUE;
@@ -910,6 +1095,99 @@ int update_xconfig(Options *op, XConfigPtr config)
 
 
 
+/*
+ * get_xserver_in_use() - try to determine which X server is in use
+ * (XFree86, Xorg); also determine if the X server supports the
+ * Extension section of the X config file; support for the "Extension"
+ * section was added between X.Org 6.7 and 6.8.
+ *
+ * Some of the parsing here mimics what is done in the
+ * check_for_modular_xorg() function in nvidia-installer
+ */
+
+#define NV_LINE_LEN 1024
+#define EXTRA_PATH "/bin:/usr/bin:/sbin:/usr/sbin:/usr/X11R6/bin:/usr/bin/X11"
+#define VERSION_FORMAT "X Protocol Version %d, Revision %d, Release %d.%d"
+
+static void get_xserver_in_use(Options *op)
+{
+#if defined(NV_SUNOS)    
+
+    /*
+     * Solaris x86/x64 always uses X.Org 6.8 or higher, atleast as far
+     * as the NVIDIA X driver is concerned
+     */
+    
+    op->gop.xserver = X_IS_XORG;
+    op->supports_extension_section = TRUE;
+    
+#else
+    
+    FILE *stream = NULL;
+    int xserver = -1;
+    int dummy, len, release_major, release_minor;
+    char *cmd, *ptr, *ret;
+    
+    op->supports_extension_section = FALSE;
+    
+    /* run `X -version` with a PATH that hopefully includes the X binary */
+    
+    cmd = xconfigStrcat("PATH=", op->gop.x_project_root, ":",
+                        EXTRA_PATH, ":$PATH X -version 2>&1", NULL);
+    
+    if ((stream = popen(cmd, "r"))) {
+        char buf[NV_LINE_LEN];
+        
+        /* read in as much of the input as we can fit into the buffer */
+        
+        ptr = buf;
+
+        do {
+            len = NV_LINE_LEN - (ptr - buf) - 1;
+            ret = fgets(ptr, len, stream);
+            ptr = strchr(ptr, '\0');
+        } while ((ret != NULL) && (len > 1));
+        
+        /* Check if this is an XFree86 release */
+        
+        if (strstr(buf, "XFree86 Version") != NULL) {
+            xserver = X_IS_XF86;
+            op->supports_extension_section = FALSE;
+        } else {
+            xserver = X_IS_XORG;
+            if ((ptr = strstr(buf, "X Protocol Version")) != NULL &&
+                sscanf(ptr, VERSION_FORMAT, &dummy, &dummy,
+                       &release_major, &release_minor) == 4) {
+                
+                if ((release_major > 6) ||
+                    ((release_major == 6) && (release_minor >= 8))) {
+                    op->supports_extension_section = TRUE; 
+                }
+            }
+        } 
+    }
+    /* Close the popen()'ed stream. */
+    pclose(stream);
+    free(cmd);
+
+    if (xserver == -1) {
+        char *xorgpath;
+
+        xorgpath = xconfigStrcat(op->gop.x_project_root, "/bin/Xorg", NULL);
+        if (access(xorgpath, F_OK)==0) {
+            xserver = X_IS_XORG;
+        } else {
+            xserver = X_IS_XF86;
+        }
+        free(xorgpath);
+    }
+    
+    op->gop.xserver=xserver;
+#endif
+} /* get_xserver_in_use */
+
+
+
 /*
  * main program entry point
  *
@@ -927,18 +1205,9 @@ int main(int argc, char *argv[])
     int ret;
     XConfigPtr config = NULL;
     
-
-    /* Load defaults */
-
-    op = load_default_options();
-    if (!op) {
-        fprintf(stderr, "\nOut of memory error.\n\n");
-        return 1;
-    }
-
     /* parse the commandline */
 
-    parse_commandline(op, argc, argv);
+    op = parse_commandline(argc, argv);
     
     /*
      * first, check for any of special options that cause us to exit
@@ -964,7 +1233,7 @@ int main(int argc, char *argv[])
         return (ret ? 0 : 1);
     }
  
-    if (op->extract_edids_from_log) {
+    if (op->extract_edids_from_file) {
         ret = extract_edids(op);
         return (ret ? 0 : 1);
     }
@@ -990,7 +1259,7 @@ int main(int argc, char *argv[])
     /*
      * Get which X server is in use: Xorg or XFree86
      */
-    xconfigGetXServerInUse(&op->gop);
+    get_xserver_in_use(op);
     
     /*
      * if we failed to find the system's config file, generate a new
